from .sum import copy
from .edge import Edge
from .variable import Variable
from .constant import Constant
from .vstack import split
from proximal.utils.timings_log import TimingsLog
import copy as cp
from collections import defaultdict
import numpy as np
from scipy.sparse.linalg import lsqr, LinearOperator, eigs

class CompGraph(object):
    """A computation graph representing a composite lin op.
    """
    def __init__(self, end, implem=None):
        self.orig_end = end
        self.end = cp.copy(end)
        self.shape = self.end.shape
        # Construct via graph traversal.
        self.nodes = []
        self.edges = []
        self.constants = []
        self.input_edges = {}
        self.output_edges = {}
        new_vars = []
        # Assumes all nodes have at most one output.
        ready = [self.end]
        while len(ready) > 0:
            curr = ready.pop(0)
            if isinstance(curr, Variable):
                new_vars.append(curr)
            # Zero out constants.
            self.nodes.append(curr)
            input_edges = []
            for node in curr.input_nodes:
                # Zero out constants.
                if isinstance(node, Constant):
                    node = Constant(np.zeros(curr.shape))
                    self.constants.append(node)
                else:
                    node = cp.copy(node)
                    # Default implementation.
                    if implem is not None:
                        node.implem = implem
                ready.append(node)
                edge = Edge(node, curr, node.shape)
                input_edges.append(edge)
                self.output_edges[node] = [edge]

            self.edges += input_edges
            self.input_edges[curr] = input_edges

        # Make copy node for each variable.
        old_vars = self.orig_end.variables()
        id2copy = {}
        for var in old_vars:
            copy_node = copy(var.shape, implem=implem)
            id2copy[var.uuid] = copy_node
            self.output_edges[copy_node] = []
            self.nodes.append(copy_node)

        # Replace variables with copy nodes in graph.
        for var in new_vars:
            copy_node = id2copy[var.uuid]
            output_edge = self.output_edges[var][0]
            output_node = output_edge.end
            edge = Edge(copy_node, output_node, var.shape)
            self.edges.append(edge)
            self.output_edges[copy_node].append(edge)
            idx = self.input_edges[output_node].index(output_edge)
            self.input_edges[output_node][idx] = edge

        # Record information about variables.
        self.input_size = sum([var.size for var in old_vars])
        self.output_size = self.end.size
        self.var_info = {}

        # Make single split node as start node.
        copy_nodes = []
        offset = 0
        for key, val in id2copy.items():
            copy_nodes.append(val)
            self.var_info[key] = offset
            offset += val.size

        self.start = split(copy_nodes, implem=implem)
        self.nodes.append(self.start)
        split_outputs = []
        for copy_node in copy_nodes:
            edge = Edge(self.start, copy_node, copy_node.shape)
            split_outputs.append(edge)
            self.input_edges[copy_node] = [edge]

        self.edges += split_outputs
        self.output_edges[self.start] = split_outputs
        # A record of timings.
        self.forward_log = TimingsLog(self.nodes + [self])
        self.adjoint_log = TimingsLog(self.nodes + [self])

    def get_inputs(self, node):
        """Returns the input data for a node.
        """
        return [e.data for e in self.input_edges[node]]

    def get_outputs(self, node):
        """Returns the output data for a node.
        """
        return [e.data for e in self.output_edges[node]]

    def forward(self, x, y):
        """Evaluates the forward composition.

        Reads from x and writes to y.
        """
        def forward_eval(node):
            if node is self.start:
                inputs = [x]
            else:
                inputs = self.get_inputs(node)

            if node is self.end:
                outputs = [y]
            else:
                outputs = self.get_outputs(node)
            # Run forward op and time it.
            self.forward_log[node].tic()
            node.forward(inputs, outputs)
            self.forward_log[node].toc()
        # Evaluate forward graph and time it.
        self.forward_log[self].tic()
        self.traverse_graph(forward_eval, True)
        self.forward_log[self].toc()

    def adjoint(self, u, v):
        """Evaluates the adjoint composition.

        Reads from u and writes to v.
        """
        def adjoint_eval(node):
            if node is self.end:
                outputs = [u]
            else:
                outputs = self.get_outputs(node)

            if node is self.start:
                inputs = [v]
            else:
                inputs = self.get_inputs(node)
            # Run adjoint op and time it.
            self.adjoint_log[node].tic()
            node.adjoint(outputs, inputs)
            self.adjoint_log[node].toc()
        # Evaluate adjoint graph and time it.
        self.adjoint_log[self].tic()
        self.traverse_graph(adjoint_eval, False)
        self.adjoint_log[self].toc()

    def traverse_graph(self, node_fn, forward):
        """Traverse the graph and apply the given function at each node.

           forward: Traverse in standard or reverse order?
           node_fn: Function to evaluate on each node.
        """
        ready = []
        eval_map = defaultdict(int)
        if forward:
            ready.append(self.start)
            # Constant nodes are leaves as well.
            ready += self.constants
        else:
            ready.append(self.end)
        while len(ready) > 0:
            curr = ready.pop()
            # Evaluate the given function on curr.
            node_fn(curr)
            eval_map[curr] += 1
            if forward:
                child_edges = self.output_edges.get(curr, [])
            else:
                child_edges = self.input_edges.get(curr, [])

            # If each input has visited the node, it is ready.
            for edge in child_edges:
                if forward:
                    node = edge.end
                else:
                    node = edge.start
                eval_map[node] += 1
                if forward:
                    node_inputs_count = len(self.input_edges[node])
                else:
                    node_inputs_count = len(self.output_edges[node])

                if (eval_map[node] == node_inputs_count):
                    ready.append(node)

    def norm_bound(self, final_output_mags):
        """Returns fast upper bound on ||K||.

        Parameters
        ----------
        final_output_mags : list
            Place to store final output magnitudes.
        """
        def node_norm_bound(node):
            # Read input magnitudes off edges.
            if node is self.start:
                input_mags = [1]
            else:
                input_mags = [e.mag for e in self.input_edges[node]]

            # If a node doesn't support norm_bound, propagate that.
            if NotImplemented in input_mags:
                output_mag = NotImplemented
            else:
                output_mag = node.norm_bound(input_mags)

            if node is self.end:
                final_output_mags[0] = output_mag
            else:
                for idx, e in enumerate(self.output_edges[node]):
                    e.mag = output_mag

        self.traverse_graph(node_norm_bound, True)

    def update_vars(self, val):
        """Map sections of val to variables.
        """
        for var in self.orig_end.variables():
            offset = self.var_info[var.uuid]
            var.value = np.reshape(val[offset:offset+var.size], var.shape)
            offset += var.size

    def __str__(self):
        return self.__class__.__name__

def est_CompGraph_norm(K, tol=1e-3, try_fast_norm=True):
    """Estimates operator norm for L = ||K||.

    Parameters
    ----------
    tol : float
        Accuracy of estimate if not trying for upper bound.
    try_fast_norm : bool
        Whether to try for a fast upper bound.

    Returns
    -------
    float
        Estimate of ||K||.
    """
    if try_fast_norm:
        output_mags = [NotImplemented]
        K.norm_bound(output_mags)
        if NotImplemented not in output_mags:
            return output_mags[0]

    input_data = np.zeros(K.input_size)
    output_data = np.zeros(K.output_size)

    def KtK(x):
        K.forward(x, output_data)
        K.adjoint(output_data, input_data)
        return input_data

    #Define linear operator
    A = LinearOperator((K.input_size, K.input_size),
                        KtK, KtK)

    Knorm = np.sqrt( eigs(A, k=1, M=None, sigma=None, which='LM', tol=tol)[0].real )
    return np.float(Knorm)



from .lin_op import LinOp
import numpy as np
from proximal.utils.utils import *
from proximal.halide.halide import *

class grad(LinOp):
    """
    gradient operation. can be defined for different dimensions.
    default is n-d gradient.
    """
    def __init__(self, arg, dims=None, implem = None):

        if dims is not None:
            self.dims = dims
        else:
            self.dims = len(arg.shape) #Full n-d gradient

        shape = arg.shape + (self.dims,)

        #Temp array for halide
        self.tmpfwd = None
        self.tmpadj = None
        if len(arg.shape) in [2,3] and self.dims == 2:
            self.tmpfwd = np.zeros( (arg.shape[0], arg.shape[1],
                                     arg.shape[2] if (len(arg.shape) > 2) else 1, 2) ,
                                    dtype=np.float32, order='FORTRAN');

            self.tmpadj = np.zeros( (arg.shape[0], arg.shape[1],
                                     arg.shape[2] if (len(arg.shape) > 2) else 1) ,
                                    dtype=np.float32, order='FORTRAN');

        super(grad, self).__init__([arg], shape, implem)

    def forward(self, inputs, outputs):
        """The forward operator for n-d gradients.

        Reads from inputs and writes to outputs.
        """

        if self.implementation == Impl['halide'] and \
            (len(self.shape) == 3 or len(self.shape) == 4) and self.dims == 2:
            #Halide implementation
            if len(self.shape) == 3:
                tmpin = np.asfortranarray( (inputs[0][..., np.newaxis]).astype(np.float32) )
            else:
                tmpin = np.asfortranarray( (inputs[0]).astype(np.float32) )

            Halide('A_grad.cpp').A_grad( tmpin, self.tmpfwd ) #Call
            np.copyto(outputs[0], np.reshape( self.tmpfwd, self.shape ) )

        else:

            #Input
            f = inputs[0]

            #Build up index for shifted array
            ss = f.shape;
            stack_arr = ()
            for j in range(self.dims):

                #Add grad for this dimension (same as index)
                il = ()
                for i in range(len(ss)):
                    if i == j:
                        il += np.index_exp[ np.r_[1:ss[j],ss[j] - 1] ]
                    else:
                        il += np.index_exp[:]

                fgrad_j = f[ il ] - f;
                stack_arr += (fgrad_j,)

            #Stack all grads as new dimension
            np.copyto( outputs[0], np.stack( stack_arr, axis=-1 ) )

    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """

        if self.implementation == Impl['halide'] and \
            (len(self.shape) == 3 or len(self.shape) == 4) and self.dims == 2:

            #Halide implementation
            if len(self.shape) == 3:
                tmpin = np.asfortranarray( np.reshape(inputs[0],
                    (self.shape[0], self.shape[1],1,2)).astype(np.float32) )
            else:
                tmpin = np.asfortranarray( inputs[0].astype(np.float32) )

            Halide('At_grad.cpp').At_grad( tmpin, self.tmpadj ) #Call
            np.copyto(outputs[0], np.reshape( self.tmpadj, self.shape[:-1] ) )

        else:

            #Compute comparison (Negative divergence)
            f = inputs[0]

            outputs[0].fill(0.0)
            for j in [0,1]:

                #Get component
                fj = f[...,j]
                ss = fj.shape

                #Add grad for this dimension (same as index)
                istart = ()
                ic = ()
                iend_out = ()
                iend_in = ()
                for i in range(len(ss)):
                    if i == j:
                        istart += np.index_exp[0]
                        ic += np.index_exp[ np.r_[0,0:ss[j]-1] ]
                        iend_out += np.index_exp[-1]
                        iend_in += np.index_exp[-2]
                    else:
                        istart += np.index_exp[:]
                        ic += np.index_exp[:]
                        iend_in += np.index_exp[:]
                        iend_out += np.index_exp[:]

                #Do the grad operation for dimension j
                fd = fj - fj[ic]
                fd[istart] = fj[istart]
                fd[iend_out] = -fj[iend_in]

                outputs[0] += (-fd)

    def get_dims(self):
        """Return the dimensinonality of the gradient
        """
        return self.dims

    def norm_bound(self, input_mags):
        """Gives an upper bound on the magnitudes of the outputs given inputs.

        Parameters
        ----------
        input_mags : list
            List of magnitudes of inputs.

        Returns
        -------
        float
            Magnitude of outputs.
        """
        # 1D gradient operator has spectral norm = 2.
        # ND gradient is permutation of stacked grad in axis 0, axis 1, etc.
        # so norm is 2*sqrt(dims)
        return 2*np.sqrt(self.dims)*input_mags[0]



from __future__ import division
import abc
import numpy as np
from proximal.utils import Impl

class ProxFn(object):
    """Represents alpha*f(beta*x - b) + <c,x> + gamma*<x,x> + d
    """
    __metaclass__ = abc.ABCMeta
    def __init__(self, lin_op, alpha=1.0, beta=1.0, b=0.0, c=0.0,
        gamma=0.0, d=0.0, implem=None):
        # Error checking.
        for elem, name in zip([b,c], ["b","c"]):
            if not (np.isscalar(elem) or elem.shape == lin_op.shape):
                raise Exception("Invalid dimensions of %s." % name)
        for elem, name in zip([alpha, gamma, d], ["alpha", "gamma"]):
            if not np.isscalar(elem) or elem < 0:
                raise Exception("%s must be a nonnegative scalar." % name)
        for elem, name in zip([beta, d], ["beta", "d"]):
            if not np.isscalar(elem):
                raise Exception("%s must be a scalar." % name)

        self.implem_key = implem
        self.implementation = Impl['numpy']
        if implem is not None:
            self.set_implementation(implem)

        self.lin_op = lin_op
        self.alpha = float(alpha)
        self.beta = float(beta)
        self.b = b
        self.c = c
        if np.isscalar(b):
            self.b = b*np.ones(self.lin_op.shape)
        if np.isscalar(c):
            self.c = c*np.ones(self.lin_op.shape)
        self.gamma = float(gamma)
        self.d = float(d)
        self.init_tmps()
        super(ProxFn, self).__init__()

    def set_implementation(self, im):
        if im in Impl.values():
            self.implementation = im
        elif im in Impl.keys():
            self.implementation = Impl[im]
        else:
            raise Exception("Invalid implementation.")

        return self.implementation

    def implementation(self, im):
        return self.implementation

    def variables(self):
        """Return a list of the variables in the problem.
        """
        return self.lin_op.variables()

    def init_tmps(self):
        """Initialize temporary variables for _prox method.
        """
        pass

    @abc.abstractmethod
    def _prox(self, rho, v, *args, **kwargs):
        """The prox function for a specific atom.
        """
        return NotImplemented

    def prox(self, rho, v, *args, **kwargs):
        """Wrapper on the prox function to handle alpha, etc.
           It is here the iteration for debug purposese etc.
        """
        rho_hat = (rho+2*self.gamma)/(self.alpha*self.beta**2)
        # vhat = (rho*v - c)*beta/(rho + 2*gamma) - b
        # Modify v in-place. This is important for the Python to be performant.
        v *= rho
        v -= self.c
        v *= self.beta/(rho + 2*self.gamma)
        v -= self.b
        xhat = self._prox(rho_hat, v, *args, **kwargs)
        # x = (xhat + b)/beta
        # Modify result in-place.
        xhat += self.b
        xhat /= self.beta
        return xhat

    @abc.abstractmethod
    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        return NotImplemented

    def eval(self, v):
        """Evaluate the function on v.
        """
        return self.alpha*self._eval(self.beta*v - self.b) + \
        np.sum(self.c*v) + self.gamma*np.square(v).sum() + self.d

    @property
    def value(self):
        return self.eval(self.lin_op.value)

    def __str__(self):
        """Default to string is name of class.
        """
        return self.__class__.__name__

    def __add__(self, other):
        """ProxFn + ProxFn(s).
        """
        if isinstance(other, ProxFn):
            return [self, other]
        elif type(other) == list:
            return [self] + other
        else:
            return NotImplemented

    def __radd__(self, other):
        """Called for list + ProxFn.
        """
        if type(other) == list:
            return other + [self]
        else:
            raise NotImplemented

    def __mul__(self, other):
        """ProxFn * Number.
        """
        # Can only multiply by scalar constants.
        if np.isscalar(other) and other > 0:
            return self.copy(alpha=self.alpha*other)
        else:
            raise TypeError("Can only multiply by a positive scalar.")

    def __rmul__(self, other):
        """Called for Number * ProxFn.
        """
        return self * other

    def __div__(self, other):
        """Called for ProxFn/Number.
        """
        return (1./other)*self

    def copy(self, lin_op=None, **kwargs):
        """Returns a shallow copy of the object.

        Used to reconstruct an object tree.

        Parameters
        ----------
        args : list, optional
            The arguments to reconstruct the object. If args=None, use the
            current args of the object.

        Returns
        -------
        Expression
        """
        if lin_op is None:
            lin_op = self.lin_op
        data = self.get_data()
        curr_args = {'alpha':self.alpha,
                     'beta':self.beta,
                     'gamma':self.gamma,
                     'c':self.c,
                     'b':self.b,
                     'd':self.d,
                     'implem':self.implem_key}
        for key in curr_args.keys():
            if key not in kwargs:
                kwargs[key] = curr_args[key]
        return type(self)(lin_op, *data, **kwargs)

    def get_data(self):
        """Returns info needed to reconstruct the object besides the args.

        Returns
        -------
        list
        """
        return []



from .prox_fn import ProxFn
from proximal.lin_ops import CompGraph, vstack, mul_elemwise
import numpy as np
from proximal.utils.utils import *
from scipy.sparse.linalg import lsqr, LinearOperator
from proximal.halide.halide import *

class sum_squares(ProxFn):
    """The function ||x||_2^2.
    """
    def absorb_params(self):
        """Returns an equivalent sum_squares with alpha = 1.0,
           gamma = 0, and c = 0.
        """
        new_beta = np.sqrt(self.alpha*self.beta**2 + self.gamma)
        new_b = (self.alpha*self.beta*self.b - self.c/2)/new_beta
        return sum_squares(self.lin_op, beta=new_beta, b=new_b)

    def _prox(self, rho, v, *args, **kwargs):
        """x = rho/(2+rho)*v.
        """
        v *= rho/(2+rho)
        return v

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        return np.square(v).sum()

class weighted_sum_squares(sum_squares):
    """The function ||W.*x||_2^2.
    """
    def __init__(self, lin_op, weight, **kwargs):
        self.weight = weight
        super(weighted_sum_squares, self).__init__(lin_op, **kwargs)

    def absorb_params(self):
        """Returns an equivalent sum_squares with alpha = 1.0,
           gamma = 0, and c = 0.
        """
        new_lin_op = mul_elemwise(self.weight, self.lin_op)
        new_b = mul_elemwise(self.weight, self.b).value
        return sum_squares(new_lin_op, alpha=self.alpha,
            beta=new_beta, b=new_b, c=self.c, gamma=self.gamma).absorb_params()

    def _prox(self, rho, v, *args, **kwargs):
        """x = (rho/weight)/(2+(rho/weight))*v.
        """
        rho_vec = rho/np.square(self.weight[self.weight != 0])
        v[self.weight != 0] *= rho_vec/(2 + rho_vec)
        return v

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        return super(weighted_sum_squares, self)._eval(self.weight*v)

    def get_data(self):
        """Returns info needed to reconstruct the object besides the args.

        Returns
        -------
        list
        """
        return [self.weight]

class least_squares(sum_squares):
    """The function ||K*x||_2^2.

       Here K is a computation graph (vector to vector lin op).
    """
    def __init__(self, lin_op, offset, diag=None, freq_diag=None,
                 freq_dims=None, implem = Impl['numpy'], **kwargs):
        self.K = CompGraph(lin_op)
        self.offset = offset
        self.diag = diag
        # TODO: freq diag is supposed to be True/False. What is going on below?
        self.freq_diag = freq_diag
        self.orig_freq_diag = freq_diag
        self.freq_dims = freq_dims
        self.orig_freq_dims = freq_dims
        #Get shape for frequency inversion var
        if self.freq_diag is not None:
            if len(self.K.orig_end.variables()) > 1:
                raise Exception("Diagonal frequency inversion supports only one var currently.")

            self.freq_shape = self.K.orig_end.variables()[0].shape
            self.freq_diag = np.reshape( self.freq_diag, self.freq_shape )
            if implem == Impl['halide'] and \
                ( len(self.freq_shape) == 2 or (len(self.freq_shape) == 2 and self.freq_dims == 2) ):
                print "hello"
                #TODO: FIX REAL TO IMAG
                hsize = self.freq_shape if len(self.freq_shape) == 3 else (self.freq_shape[0], self.freq_shape[1], 1)
                hsizehalide = ( (hsize[0] + 1)/2 + 1, hsize[1], hsize[2], 2)

                self.hsizehalide = hsizehalide
                self.ftmp_halide = np.zeros( hsizehalide, dtype=np.float32, order='F' )
                self.ftmp_halide_out = np.zeros( hsize, dtype=np.float32, order='F' )
                self.freq_diag = np.reshape( self.freq_diag[0:hsizehalide[0], ...], hsizehalide[0:3] )

        super(least_squares, self).__init__(lin_op, implem = implem, **kwargs)

    def get_data(self):
        """Returns info needed to reconstruct the object besides the args.

        Returns
        -------
        list
        """
        return [self.offset, self.diag, self.orig_freq_diag, self.orig_freq_dims]

    def _prox(self, rho, v, b=None, lin_solver="cg", *args, **kwargs):
        """x = argmin_x ||K*x - self.offset - b||_2^2 + (rho/2)||x-v||_2^2.
        """
        if b is None:
            offset = self.offset
        else:
            offset = self.offset + b
        return self.solve(offset, rho=rho, v=v, lin_solver=lin_solver, *args, **kwargs)

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        Kv = np.zeros(self.K.output_size)
        self.K.forward(v.ravel(), Kv)
        return super(least_squares, self)._eval(Kv - self.offset)

    def solve(self, b, rho=None, v=None, lin_solver="lsqr", *args, **kwargs):
        # KtK Operator is diagonal
        if self.diag is not None:

            Ktb = np.zeros(self.K.input_size)
            self.K.adjoint(b, Ktb)
            if rho is None:
                Ktb /= self.diag
            else:
                Ktb += (rho/2.) * v
                Ktb /= (self.diag + rho/2.)

            return Ktb

        # KtK operator is diagonal in frequency domain.
        elif self.freq_diag is not None:
            Ktb = np.zeros(self.K.input_size)
            self.K.adjoint(b, Ktb)

            #Frequency inversion
            if self.implementation == Impl['halide'] and \
                ( len(self.freq_shape) == 2 or (len(self.freq_shape) == 2 and self.freq_dims == 2) ):

                Halide('fft2_r2c.cpp').fft2_r2c( np.asfortranarray( np.reshape(Ktb.astype(np.float32), self.freq_shape) ), 0, 0, self.ftmp_halide)

                Ktb = 1j*self.ftmp_halide[...,1]
                Ktb += self.ftmp_halide[...,0]

                if rho is None:
                    Ktb /= self.freq_diag
                else:
                    Halide('fft2_r2c.cpp').fft2_r2c( np.asfortranarray( np.reshape(v.astype(np.float32), self.freq_shape) ), 0, 0, self.ftmp_halide)

                    vhat = self.ftmp_halide[...,0] + 1j*self.ftmp_halide[...,1]
                    Ktb *= 1.0/rho
                    Ktb += vhat
                    Ktb /= (1.0/rho * self.freq_diag + 1.0)

                #Do inverse tranform
                Ktb = np.asfortranarray( np.stack( (Ktb.real, Ktb.imag ), axis=-1 ) )
                Halide('ifft2_c2r.cpp').ifft2_c2r(Ktb, self.ftmp_halide_out)

                return self.ftmp_halide_out.ravel()

            else:

                #General frequency inversion
                Ktb = fftd(np.reshape(Ktb, self.freq_shape), self.freq_dims)

                if rho is None:
                    Ktb /= self.freq_diag
                else:
                    Ktb *= 2.0/rho
                    Ktb += fftd(np.reshape(v,self.freq_shape), self.freq_dims)
                    Ktb /= (2.0/rho * self.freq_diag + 1.0)

                return (ifftd(Ktb, self.freq_dims).real).ravel()

        elif lin_solver == "lsqr":
            return self.solve_lsqr(b, rho, v, *args, **kwargs)
        elif lin_solver == "cg":
            return self.solve_cg(b, rho, v, *args, **kwargs)
        else:
            raise Exception("Unknown least squares solver.")

    def solve_lsqr(self, b, rho=None, v=None, x_init=None, options=None):
        """Solve ||K*x - b||^2_2 + (rho/2)||x-v||_2^2.
        """

        #Add additional linear terms for the rho terms
        sizev = 0
        if rho is not None:
            vf = v.flatten() * np.sqrt(rho/2.0)
            sizeb = self.K.input_size
            sizev = np.prod(v.shape)
            b = np.hstack( (b, vf) )

        input_data = np.zeros(self.K.input_size)
        output_data = np.zeros(self.K.output_size + sizev)

        def matvec(x, output_data):
            if rho is None:
                #Traverse compgraph
                self.K.forward(x, output_data )
            else:
                #Compgraph and additional terms
                self.K.forward(x, output_data[0:0 + sizeb] )
                np.copyto( output_data[sizeb:sizeb + sizev], x * np.sqrt(rho/2.0) )

            return output_data

        def rmatvec(y, input_data):
            if rho is None:
                self.K.adjoint(y, input_data)
            else:
                self.K.adjoint(y[0:0 + sizeb], input_data)
                input_data += y[sizeb:sizeb + sizev] * np.sqrt(rho/2.0)

            return input_data

        #Define linear operator
        matvecComp = lambda x: matvec(x, output_data)
        rmatvecComp = lambda y: rmatvec(y, input_data)

        K = LinearOperator((self.K.output_size + sizev, self.K.input_size),
                            matvecComp, rmatvecComp)

        #Options
        if options is None:
            #Default options
            return lsqr(K, b)[0]
        else:
            if not isinstance(options, lsqr_options):
                raise Exception("Invalid LSQR options.")
            return lsqr(K, b, atol=options.atol, btol = options.btol, show=options.show, iter_lim=options.iter_lim)[0]

    def solve_cg(self, b, rho=None, v=None, x_init=None, options=None):
        """Solve ||K*x - b||^2_2 + (rho/2)||x-v||_2^2.
        """
        output_data = np.zeros(self.K.output_size)

        def KtK(x, r):
            self.K.forward(x, output_data)
            self.K.adjoint(output_data, r)
            if rho is not None:
                r += rho * x
            return r

        #Compute Ktb
        Ktb = np.zeros(self.K.input_size)
        self.K.adjoint(b, Ktb)
        if rho is not None:
            Ktb += rho * v

        #Options
        if options is None:
            #Default options
            options = cg_options()
        elif not isinstance(options, cg_options):
            raise Exception("Invalid CG options.")

        return cg(KtK, Ktb, options.tol, options.num_iters, options.verbose, x_init,  self.implementation)

class lsqr_options:
    def __init__(self, atol=1e-6, btol=1e-6, num_iters=50, verbose=False):
        self.atol = atol
        self.btol = btol
        self.iter_lim = num_iters
        self.show = verbose

class cg_options:
    def __init__(self, tol=1e-6, num_iters=50, verbose=False):
        self.tol = tol
        self.num_iters = num_iters
        self.verbose = verbose

def cg( KtKfun, b, tol, num_iters, verbose, x_init = None, implem = Impl['numpy'] ):

    # Solves KtK x = b with
    # KtKfun being a function that computes the matrix vector product KtK x

    #TODO: Fix halide later
    implem == Impl['numpy']

    if implem == Impl['halide']:
        output = np.array([0.0], dtype=np.float32)
        hl_norm2 = Halide('A_norm_L2.cpp', generator_name="normL2_1DImg", func="A_norm_L2_1D").A_norm_L2_1D
        hl_dot = Halide('A_dot_prod.cpp', generator_name="dot_1DImg", func="A_dot_1D").A_dot_1D

        #Temp vars
        x = np.zeros(b.shape, dtype=np.float32, order='F')
        r = np.zeros(b.shape, dtype=np.float32, order='F')
        Ap = np.zeros(b.shape, dtype=np.float32, order='F')

    else:
        #Temp vars
        x = np.zeros(b.shape)
        r = np.zeros(b.shape)
        Ap = np.zeros(b.shape)

    #Initialize x
    # Initialize everything to zero.
    if x_init is not None:
        x = x_init

    # Compute residual
    # r = b - KtKfun(x)
    KtKfun(x,r)
    r *= -1.0
    r += b

    # Do cg iterations
    if implem == Impl['halide']:
        hl_norm2(b.ravel().astype(np.float32), output)
        cg_tol = tol * output[0]
    else:
        cg_tol = tol * np.linalg.norm(b.ravel(), 2)  #Relative tol

    # CG iteration
    cg_iter = np.minimum( num_iters, np.prod(b.shape) )
    for iter in range(cg_iter):
        # Check for convergence

        if implem == Impl['halide']:
            hl_norm2(r.ravel(), output)
            normr = output[0]
        else:
            normr = np.linalg.norm(r.ravel(), 2)

        #Check for convergence
        if normr <= cg_tol :
            break

        #gamma = r'*r;
        if implem == Impl['halide']:
            hl_norm2(r.ravel(), output)
            gamma = output[0]
            gamma *= gamma
        else:
            gamma = np.dot( r.ravel().T, r.ravel() )

        #direction vector
        if iter > 0 :
            beta = gamma / gamma_1;
            p = r + beta * p;
        else:
            p = r;

        # Compute Ap
        KtKfun(p,Ap)

        #Cg update
        q = Ap

        #alpha = gamma / (p'*q);
        if implem == Impl['halide']:
            hl_dot(p.ravel(), q.ravel(), output)
            alpha = gamma / output[0]
        else:
            alpha = gamma / np.dot( p.ravel().T, q.ravel() )

        x = x + alpha * p  # update approximation vector
        r = r - alpha * q  # compute residual

        gamma_1 = gamma

        # Iterate
        if verbose :
            print "CG Iter %03d" % iter

    return x



import timeit

class TimingsEntry(object):
    """A log of the runtime for an operation.
    """
    def __init__(self, op):
        self.op = op
        self.evals = 0
        self.total_time = 0
        self.lastticstamp = None

    @property
    def avg_time(self):
        if self.evals == 0:
            return 0
        else:
            return self.total_time/self.evals

    def record_timing(self, elapsed):
        """Updates the log with the new time.
        """
        self.evals += 1
        self.total_time += elapsed

    def tic(self):
        """ Default timer
        Example: t = tic()
             ... code
             elapsed = toc(t)
             print( '{0}: {1:.4f}ms'.format(message, elapsed) )
        """

        t = timeit.default_timer()
        self.lastticstamp = t
        return t

    def toc(self):
        """ See tic f
        """

        #Last tic
        if self.lastticstamp is None:
            raise Exception('Error: Call to toc did never call tic before.')
        else:
            t = self.lastticstamp
            #Measure time in ms
            elapsed = (timeit.default_timer() - t) * 1000.0  #in ms
            # Update recrod.
            self.record_timing(elapsed)
            self.lastticstamp = None
            return elapsed

    def __str__(self):
        return "op = %s, evals = %s, total_time (ms) = %s, avg_time (ms) = %s" % (
            self.op, self.evals, self.total_time, self.avg_time)

class TimingsLog(object):
    """A log of the runtime for a set of operations.
    """
    def __init__(self, ops):
        self.ops = ops
        self.data = {}
        for op in self.ops:
            self.data[op] = TimingsEntry(op)

    def __getitem__(self, item):
        return self.data[item]

    def __str__(self):
        logs = []
        for op in self.ops:
            if self[op].evals > 0:
                logs += [str(self.data[op])]
        return '\n'.join(logs)



from .lin_op import LinOp
import numpy as np

class scale(LinOp):
    """Multiplication scale*X with a fixed scalar.
    """
    def __init__(self, scalar, arg):
        assert np.isscalar(scalar)
        self.scalar = scalar
        shape = arg.shape
        super(scale, self).__init__([arg], shape)

    def forward(self, inputs, outputs):
        """The forward operator.

        Reads from inputs and writes to outputs.
        """
        np.multiply(inputs[0], self.scalar, outputs[0])

    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """
        self.forward(inputs, outputs)

    def is_diag(self, freq=False):
        """Is the lin op diagonal (in the frequency domain)?
        """
        return self.input_nodes[0].is_diag(freq)

    def get_diag(self, freq=False):
        """Returns the diagonal representation (A^TA)^(1/2).

        Parameters
        ----------
        freq : bool
            Is the diagonal representation in the frequency domain?
        Returns
        -------
        dict of variable to ndarray
            The diagonal operator acting on each variable.
        """
        var_diags = self.input_nodes[0].get_diag(freq)
        for var in var_diags.keys():
            var_diags[var] = var_diags[var]*self.scalar
        return var_diags

    def norm_bound(self, input_mags):
        """Gives an upper bound on the magnitudes of the outputs given inputs.

        Parameters
        ----------
        input_mags : list
            List of magnitudes of inputs.

        Returns
        -------
        float
            Magnitude of outputs.
        """
        return abs(self.scalar)*input_mags[0]



from proximal.lin_ops import (CompGraph, est_CompGraph_norm, scale, Variable,
    vstack)
from proximal.prox_fns import least_squares
from proximal.utils.timings_log import TimingsLog, TimingsEntry
from invert import get_least_squares_inverse, max_diag_set
import numpy as np

def partition(prox_fns, try_diagonalize=True):
    """Divide the proxable functions into sets Psi and Omega.
    """
    # Omega must be a single function.
    # Merge quadratic functions into the x update.
    # Automatically try to split the problem.
    quad_fns = max_diag_set(prox_fns)
    split_fn = []
    omega_fns = []
    if len(quad_fns) == 0:
        for fn in prox_fns:
            if type(fn.lin_op) == Variable:
                split_fn = [fn]
                break
        omega_fns = split_fn
    else:
        # Proximal solve for:
        # G(x) + 1/(2*tau) * ||x - v||^2_2, with G containing all quadratics
        quad_ops = []
        const_terms = []
        for fn in quad_fns:
            fn = fn.absorb_params()
            quad_ops.append( fn.beta*fn.lin_op )
            const_terms.append( fn.b.flatten() )

        stacked_ops = vstack(quad_ops)
        b = np.hstack(const_terms)
        # Get optimize inverse (tries spatial and frequency diagonalization)
        x_update = get_least_squares_inverse(quad_ops, b, try_diagonalize)
        omega_fns = [x_update]

    psi_fns = [fn for fn in prox_fns if fn not in split_fn + quad_fns]
    return psi_fns, omega_fns

def solve(psi_fns, omega_fns, tau=None, sigma=None, theta=None,
          max_iters=1000, eps_abs=1e-3, eps_rel=1e-3, x0 = None,
          lin_solver="cg", lin_solver_options=None,
          try_diagonalize=True, try_fast_norm=False, scaled=True,
          metric=None, convlog=None, verbose=0):
    # Can only have one omega function.
    assert len(omega_fns) <= 1
    prox_fns = psi_fns + omega_fns
    stacked_ops = vstack([fn.lin_op for fn in psi_fns])
    K = CompGraph(stacked_ops)
    v = np.zeros(K.input_size)
    #Select optimal parameters if wanted
    if tau is None or sigma is None or theta is None:
        tau, sigma, theta = est_params_pc(K, tau, sigma, verbose, scaled, try_fast_norm)

    #Initialize
    x = np.zeros(K.input_size)
    y = np.zeros(K.output_size)
    xbar = np.zeros(K.input_size)
    u = np.zeros(K.output_size)
    z = np.zeros(K.output_size)

    if x0 is not None:
        x[:] = np.reshape( x0, K.input_size )
        K.forward(x, y);
        xbar[:] = x

    # Buffers.
    Kxbar = np.zeros(K.output_size)
    Kx = np.zeros(K.output_size)
    KTy = np.zeros(K.input_size)
    KTu = np.zeros(K.input_size)
    s = np.zeros(K.input_size)

    prev_x = x.copy()
    prev_Kx = Kx.copy()
    prev_z = z.copy()
    prev_u = u.copy()

    # Log for prox ops.
    prox_log = TimingsLog(prox_fns)
    # Time iterations.
    iter_timing = TimingsEntry("PC iteration")

    #Convergence log for initial iterate
    if convlog != None:
        K.update_vars(x)
        objval = sum([fn.value for fn in prox_fns])
        convlog.record_objective( objval )
        convlog.record_timing( 0.0 )

    for i in range(max_iters):
        iter_timing.tic()
        if convlog != None:
            convlog.tic()

        #Keep track of previous iterates
        np.copyto( prev_x, x)
        np.copyto( prev_z, z)
        np.copyto( prev_u, u)
        np.copyto( prev_Kx, Kx)

        #Compute z
        K.forward(xbar, Kxbar)
        z = y + sigma*Kxbar

        # Update y.
        offset = 0
        for fn in psi_fns:
            slc = slice(offset, offset+fn.lin_op.size, None)
            z_slc = np.reshape(z[slc], fn.lin_op.shape)

            # Moreau identity: apply and time prox.
            prox_log[fn].tic()
            y[slc] = (z_slc - sigma*fn.prox(sigma, z_slc/sigma, i)).flatten()
            prox_log[fn].toc()
            offset += fn.lin_op.size
        y[offset:] = 0

        #Update x
        K.adjoint(y, KTy)
        x -= tau*KTy

        if len(omega_fns) > 0:
            xtmp = np.reshape(x, omega_fns[0].lin_op.shape)
            x[:] = omega_fns[0].prox(1.0/tau, xtmp, x_init=prev_x,
                                     lin_solver=lin_solver, options=lin_solver_options).flatten()

        #Update xbar
        np.copyto(xbar, x)
        xbar += theta*(x - prev_x)

        #Convergence log
        if convlog != None:
            convlog.toc()
            K.update_vars(x)
            objval = sum([fn.value for fn in prox_fns])
            convlog.record_objective( objval )

        """ Old convergence check
        #Very basic convergence check.
        r_x = np.linalg.norm(x - prev_x)
        r_xbar = np.linalg.norm(xbar - prev_xbar)
        r_ybar = np.linalg.norm(y - prev_y)
        error = r_x + r_xbar + r_ybar
        """

        #Residual based convergence check
        K.forward(x, Kx)
        u = 1.0/sigma * y + theta* (Kx - prev_Kx)
        z = prev_u + prev_Kx - 1.0/sigma * y

        #Iteration order is different than
        # lin-admm (--> start checking at iteration 1)
        if i > 0:

            # Check convergence
            r = prev_Kx - z
            K.adjoint( sigma*(z - prev_z), s)
            eps_pri = np.sqrt(K.output_size)*eps_abs + eps_rel*max([np.linalg.norm(prev_Kx), np.linalg.norm(z)])

            K.adjoint(u, KTu)
            eps_dual = np.sqrt(K.input_size)*eps_abs + eps_rel*np.linalg.norm(KTu)/sigma

            #Progress
            if verbose > 0:
                #Evaluate objective only if required (expensive !)
                objstr = ''
                if verbose == 2:
                    K.update_vars(x)
                    objstr = ", obj_val = %02.03e" % sum([fn.value for fn in prox_fns])

                """ Old convergence check
                #Evaluate metric potentially
                metstr = '' if metric is None else ", {}".format( metric.message(x.copy()) )
                print "iter [%04d]:" \
                      "||x - x_prev||_2 = %02.02e " \
                      "||xbar - xbar_prev||_2 = %02.02e " \
                      "||y - y_prev||_2 = %02.02e " \
                      "SUM = %02.02e (eps=%02.03e)%s%s" \
                        % (i, r_x, r_xbar, r_ybar, error, eps, objstr, metstr)
                """

                #Evaluate metric potentially
                metstr = '' if metric is None else ", {}".format( metric.message(v) )
                print "iter %d: ||r||_2 = %.3f, eps_pri = %.3f, ||s||_2 = %.3f, eps_dual = %.3f%s%s" % (
                    i, np.linalg.norm(r), eps_pri, np.linalg.norm(s), eps_dual, objstr, metstr)

            iter_timing.toc()
            if np.linalg.norm(r) <= eps_pri and np.linalg.norm(s) <= eps_dual:
                break

        else:
            iter_timing.toc()

        """ Old convergence check
        if error <= eps:
            break
        """

    # Print out timings info.
    if verbose > 0:
        print iter_timing
        print "prox funcs:"
        print prox_log
        print "K forward ops:"
        print K.forward_log
        print "K adjoint ops:"
        print K.adjoint_log

    # Assign values to variables.
    K.update_vars(x)

    # Return optimal value.
    return sum([fn.value for fn in prox_fns])

def est_params_pc(K, tau=None, sigma=None, verbose=True, scaled=False, try_fast_norm=False):

    #Select theta constant and sigma larger 0
    theta = 1.0
    sigma = 1.0 if sigma is None else sigma

    #Estimate Lipschitz constant and comput tau
    if scaled:
        L = 1
    else:
        L = est_CompGraph_norm(K, try_fast_norm)
    tau = 1.0/( sigma * L**2 )

    if verbose:
        print "Estimated params [sigma = %3.3f | tau = %3.3f | theta = %3.3f | L_est = %3.4f]" % (sigma, tau, theta, L)

    return tau, sigma, theta



from proximal.tests.base_test import BaseTest
from proximal.halide.halide import *
from proximal.utils.utils import *

import os
import numpy as np
from PIL import Image

class TestHalide(BaseTest):

    def test_compile(self):
        """Test compilation and run.
        """

        #Force recompilation of conv
        Halide('A_conv.cpp', recompile=True, verbose=False) 

    def test_slicing(self):
        """Test slicing  over numpy arrays in halide.
        """
        #Load image
        testimg_filename = os.path.join( os.path.dirname(os.path.realpath(__file__)), 'data', 'angela.jpg' )
        np_img = im2nparray( Image.open(testimg_filename) )  
        np_img = np.asfortranarray( np.tile(np_img[..., np.newaxis], (1,1,1,3) ) )

        #Test problem
        output = np.zeros_like(np_img);
        mask = np.asfortranarray(  np.random.randn( *list(np_img.shape[0:3]) ).astype(np.float32) )
        mask = np.maximum( mask, 0.)
                
        for k in range( np_img.shape[3]):
            Halide('A_mask.cpp').A_mask( np.asfortranarray(np_img[:,:,:,k]), mask, output[:,:,:,k]) #Call
        
        output_ref = np.zeros_like(np_img);
        for k in range( np_img.shape[3]):
            output_ref[:,:,:,k] = mask * np_img[:,:,:,k]

        self.assertItemsAlmostEqual(output, output_ref)


from setuptools import setup

setup(
    name='proximal',
    version='0.1.1', 
    packages=['proximal',
              'proximal.prox_fns',
              'proximal.lin_ops',
              'proximal.algorithms',
              'proximal.utils',
              'proximal.halide',
              'proximal.tests',
              'proximal.tests.data'],
    package_dir = {'proximal':'proximal'},
    package_data = {'proximal.tests.data': ['angela.jpg'],
                    'proximal.halide': ['src/*.cpp', 'src/core/*', 'src/external/*', 'src/fft/*']},
    install_requires=["numpy >= 1.9",
                      "scipy >= 0.15",
                      "cv2",
                      "PIL"],
    use_2to3=True,
)



from .lin_op import LinOp
import numpy as np

class Constant(LinOp):
    """A constant.
    """
    def __init__(self, value):
        if np.isscalar(value):
            value = np.array(value)
        self._value = value
        super(Constant, self).__init__([], value.shape)

    def variables(self):
        return []

    def constants(self):
        return [self]

    def forward(self, inputs, outputs):
        """The forward operator.

        Reads from inputs and writes to outputs.
        """
        np.copyto(outputs[0], self.value)

    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """
        pass

    @property
    def value(self):
        return self._value

    def is_diag(self, freq=False):
        """Is the lin op diagonal (in the frequency domain)?
        """
        return True

    def get_diag(self, freq=False):
        """Returns the diagonal representation (A^TA)^(1/2).

        Parameters
        ----------
        freq : bool
            Is the diagonal representation in the frequency domain?
        Returns
        -------
        dict of variable to ndarray
            The diagonal operator acting on each variable.
        """
        return {}

    def norm_bound(self, input_mags):
        """Gives an upper bound on the magnitudes of the outputs given inputs.

        Parameters
        ----------
        input_mags : list
            List of magnitudes of inputs.

        Returns
        -------
        float
            Magnitude of outputs.
        """
        return 0.0



from .prox_fn import ProxFn
import numpy as np
from proximal.utils.utils import *
from proximal.halide.halide import *

class norm1(ProxFn):
    """The function ||x||_1.
    """

    def init_tmps(self):
        """Initialize temporary variables for _prox method.
        """
        self.v_sign = np.zeros(self.lin_op.shape, dtype=float)
        self.tmpout = np.zeros(self.lin_op.shape, dtype=np.float32, order='F')

    def _prox(self, rho, v, *args, **kwargs):
        """x = sign(v)*(|v| - 1/rho)_+
        """

        if self.implementation == Impl['halide'] and (len(self.lin_op.shape) in [2,3,4]):
            #Halide implementation
            tmpin = np.asfortranarray(v.astype(np.float32))
            Halide('prox_L1.cpp').prox_L1(tmpin, 1./rho, self.tmpout)
            np.copyto(v, self.tmpout)

        else:
            #Numpy implementation
            np.sign(v, self.v_sign)
            np.absolute(v, v)
            v -= 1./rho
            np.maximum(v, 0, v)
            v *= self.v_sign

        return v

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        return np.linalg.norm(v.ravel(), 1)

class weighted_norm1(norm1):
    """The function ||W.*x||_1.
    """
    def __init__(self, lin_op, weight, **kwargs):
        self.weight = weight
        super(weighted_norm1, self).__init__(lin_op, **kwargs)

    def _prox(self, rho, v, *args, **kwargs):
        """x = sign(v)*(|v| - |W|/rho)_+
        """
        idxs = self.weight != 0
        np.sign(v[idxs], self.v_sign[idxs])
        np.absolute(v[idxs], v[idxs])
        v[idxs] -= np.absolute(self.weight[idxs])/rho
        np.maximum(v[idxs], 0, v[idxs])
        v[idxs] *= self.v_sign[idxs]
        return v

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        return super(weighted_norm1, self)._eval(self.weight*v)

    def get_data(self):
        """Returns info needed to reconstruct the object besides the args.

        Returns
        -------
        list
        """
        return [self.weight]



from .lin_op import LinOp
import numpy as np
import cv2

from proximal.halide.halide import *
from proximal.utils.utils import *

class warp(LinOp):
    """Warp using a homography.
    """
    def __init__(self, arg, H, implem = None ):
        self.H = H.copy()

        #Compute inverse
        self.Hinv = np.zeros(H.shape)
        if len(H.shape) > 2 :
            for j in range(self.H.shape[2]):
                self.Hinv[:,:,j] = np.linalg.pinv(H[:,:,j])
        else :
            self.Hinv = np.linalg.pinv(H)

        #Check for the shape
        if len(H.shape) < 2 or len(H.shape) > 3:
            raise Exception('Error, warp supports only up to 4d inputs (expects first 3 to be image).')

        #Has to have third dimension
        if len(arg.shape) != 3:
            raise Exception('Images must have third dimension')

        shape = arg.shape
        if len(H.shape) == 3:
            shape += (H.shape[2],)

        #Temp array for halide
        self.tmpfwd = np.zeros( (shape[0], shape[1], shape[2] if (len(shape) > 2) else 1, H.shape[2] if (len(H.shape) > 2) else 1),
                                dtype=np.float32, order='FORTRAN');
        self.tmpadj = np.zeros( (shape[0], shape[1], shape[2] if (len(shape) > 2) else 1), dtype=np.float32, order='FORTRAN');

        #Halide homographies
        if len(H.shape) == 2:
            self.Hf = np.asfortranarray( H[...,np.newaxis].astype(np.float32) ) #Third axis for halide
            self.Hinvf = np.asfortranarray( self.Hinv[...,np.newaxis].astype(np.float32) ) #Third axis for halide
        else:
            self.Hf = np.asfortranarray( H.astype(np.float32) )
            self.Hinvf = np.asfortranarray( self.Hinv.astype(np.float32) )

        super(warp, self).__init__([arg], shape, implem)

    def forward(self, inputs, outputs):
        """The forward operator.

        Reads from inputs and writes to outputs.
        """

        if self.implementation == Impl['halide'] :

            #Halide implementation
            tmpin = np.asfortranarray( inputs[0].astype(np.float32) )
            Halide('A_warp.cpp').A_warp( tmpin, self.Hinvf, self.tmpfwd ) #Call
            np.copyto(outputs[0], np.reshape( self.tmpfwd, self.shape ) )

        else:

            #CV2 version
            inimg = inputs[0]
            if len(self.H.shape) == 2:

                warpedInput = cv2.warpPerspective(np.asfortranarray(inimg),  self.H.T, inimg.shape[1::-1], flags=cv2.INTER_LINEAR ,
                                    borderMode=cv2.BORDER_CONSTANT, borderValue=0.) #Necessary due to array layout in opencv
                np.copyto( outputs[0], warpedInput )

            else:
                for j in range(self.H.shape[2]):
                    warpedInput = cv2.warpPerspective(np.asfortranarray(inimg), self.H[:,:,j].T, inimg.shape[1::-1], flags=cv2.INTER_LINEAR ,
                                                    borderMode=cv2.BORDER_CONSTANT, borderValue=0.) #Necessary due to array layout in opencv

                    np.copyto( outputs[0][:,:,:,j], warpedInput )

    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """


        if self.implementation == Impl['halide'] :

            #Halide implementation
            if len(self.H.shape) == 2:
                tmpin = np.asfortranarray( inputs[0][..., np.newaxis].astype(np.float32) )
            else:
                tmpin = np.asfortranarray( inputs[0].astype(np.float32) )

            Halide('At_warp.cpp').At_warp( tmpin, self.Hf, self.tmpadj ) #Call
            np.copyto(outputs[0], self.tmpadj )

        else:

            #CV2 version
            inimg = inputs[0]
            if len(self.H.shape) == 2:
                # + cv2.WARP_INVERSE_MAP
                warpedInput = cv2.warpPerspective(np.asfortranarray(inimg), self.Hinv.T, inimg.shape[1::-1], flags=cv2.INTER_LINEAR,
                                    borderMode=cv2.BORDER_CONSTANT, borderValue=0.)
                np.copyto( outputs[0], warpedInput )

            else:
                outputs[0][:] = 0.0
                for j in range(self.H.shape[2]):
                    warpedInput = cv2.warpPerspective(np.asfortranarray(inimg[:,:,:,j]), self.Hinv[:,:,j].T, inimg.shape[1::-1], flags=cv2.INTER_LINEAR,
                                                    borderMode=cv2.BORDER_CONSTANT, borderValue=0.) #Necessary due to array layout in opencv
                    outputs[0] += warpedInput

    # TODO what is the spectral norm of a warp?



import admm
import pock_chambolle as pc
import half_quadratic_splitting as hqs
import linearized_admm as ladmm
from proximal.utils.utils import Impl
from proximal.lin_ops import Variable, CompGraph, est_CompGraph_norm, vstack
import absorb
import merge
import numpy as np

NAME_TO_SOLVER = {
    "admm": admm,
    "pock_chambolle": pc,
    "pc": pc,
    "half_quadratic_splitting": hqs,
    "hqs": hqs,
    "linearized_admm": ladmm,
    "ladmm": ladmm,
}

class Problem(object):
    """An object representing a convex optimization problem.
    """
    def __init__(self, prox_fns,
                 implem=Impl['numpy'], try_diagonalize=True,
                 absorb=True, merge=True,
                 try_split=True, try_fast_norm=True, scale=True,
                 psi_fns=None, omega_fns=None,
                 lin_solver="cg", solver="pc"):
        self.prox_fns = prox_fns
        self.implem = implem
        self.try_diagonalize = try_diagonalize # Auto diagonalize?
        self.try_split = try_split # Auto partition?
        self.try_fast_norm = try_fast_norm # Fast upper bound on ||K||?
        self.scale = scale # Auto scale problem?
        self.absorb = absorb # Absorb lin ops into prox fns?
        self.merge = merge # Merge prox fns?

        # Defaults for psi and omega fns.
        # Should have psi_fns + omega_fns == prox_fns
        if psi_fns is None and omega_fns is None:
            psi_fns = []
            omega_fns = []
        elif psi_fns is None:
            psi_fns = [fn for fn in prox_fns if fn not in omega_fns]
        elif omega_fns is None:
            omega_fns = [fn for fn in prox_fns if fn not in psi_fns]
        else:
            assert set(psi_fns + omega_fns) == set(prox_fns)
        self.omega_fns = omega_fns
        self.psi_fns = psi_fns

        self.solver = solver
        self.lin_solver = lin_solver

    def set_absorb(self, absorb):
        """Try to absorb lin ops in prox fns?
        """
        self.absorb = absorb

    def set_merge(self, merge):
        """Try to merge prox fns?
        """
        self.merge = merge

    def set_automatic_frequency_split(self, freq_split):
        self.freq_split = freq_split

    def set_implementation(self, implem=Impl['numpy']):
        """Set the implementation of the lin ops and proxes.
        """
        self.implem = implem 

    def set_solver(self, solver):
        """Set the solver.
        """
        self.solver = solver

    def set_lin_solver(self, lin_solver):
        """Set solver for linear systems/least squares.
        """
        self.lin_solver = lin_solver

    def solve(self, solver=None, *args, **kwargs):
        if solver is None:
            solver = self.solver

        if len(self.omega_fns + self.psi_fns) == 0:
            prox_fns = self.prox_fns
        else:
            prox_fns = self.omega_fns + self.psi_fns
        # Absorb lin ops if desired.
        if self.absorb:
            prox_fns = absorb.absorb_all_lin_ops(prox_fns)

        # Merge prox fns.
        if self.merge:
            prox_fns = merge.merge_all(prox_fns)
        # Absorb offsets.
        prox_fns = [absorb.absorb_offset(fn) for fn in prox_fns]
        # TODO more analysis of what solver to use.
        # Short circuit with one function.
        if len(prox_fns) == 1 and type(prox_fns[0].lin_op) == Variable:
            fn = prox_fns[0]
            var = fn.lin_op
            var.value = fn.prox(0, np.zeros(fn.lin_op.shape))
            return fn.value
        elif solver in NAME_TO_SOLVER:
            module = NAME_TO_SOLVER[solver]
            if len(self.omega_fns + self.psi_fns) == 0:
                if self.try_split:
                    psi_fns, omega_fns = module.partition(prox_fns,
                                                          self.try_diagonalize)
                else:
                    psi_fns = prox_fns
                    omega_fns = []
            # Scale the problem.
            if self.scale:
                K = CompGraph(vstack([fn.lin_op for fn in psi_fns]),
                              implem=self.implem)
                Knorm = est_CompGraph_norm(K, try_fast_norm=self.try_fast_norm)
                for idx, fn in enumerate(psi_fns):
                    psi_fns[idx] = fn.copy(fn.lin_op/Knorm,
                                           beta=fn.beta*np.sqrt(Knorm),
                                           implem=self.implem)
                for idx, fn in enumerate(omega_fns):
                    omega_fns[idx] = fn.copy(beta=fn.beta/np.sqrt(Knorm),
                                             implem=self.implem)
            opt_val = module.solve(psi_fns, omega_fns,
                            lin_solver=self.lin_solver,
                            try_diagonalize=self.try_diagonalize,
                            try_fast_norm=self.try_fast_norm,
                            scaled=self.scale,
                            *args, **kwargs)
            # Unscale the variables.
            if self.scale:
                for var in self.variables():
                    var.value /= np.sqrt(Knorm)
        else:
            raise Exception("Unknown solver.")

    def variables(self):
        """Return a list of variables in the problem.
        """
        vars_ = []
        for fn in self.prox_fns:
            vars_ += fn.variables()
        return list(set(vars_))



from proximal.tests.base_test import BaseTest
from proximal.lin_ops import *
from proximal.prox_fns import *
from proximal.algorithms import *
from proximal.utils.utils import Impl
import cvxpy as cvx
import numpy as np
import scipy as sp

class TestProblem(BaseTest):

    def test_problem(self):
        """Test problem object.
        """
        X = Variable((4,2))
        B = np.reshape(np.arange(8), (4,2))*1.
        prox_fns = [norm1(X), sum_squares(X, b=B)]
        prob = Problem(prox_fns)
        # prob.partition(quad_funcs = [prox_fns[0], prox_fns[1]])
        prob.set_automatic_frequency_split(False)
        prob.set_absorb(False)
        prob.set_implementation(Impl['halide'])
        prob.set_solver('admm')
        sltn = prob.solve()

        true_X = norm1(X).prox(2, B.copy())
        self.assertItemsAlmostEqual(X.value, true_X, places=2)
        sltn = prob.solve(solver="pc")
        self.assertItemsAlmostEqual(X.value, true_X, places=2)
        sltn = prob.solve(solver="hqs", eps_rel=1e-6,
                     rho_0 = 1.0, rho_scale = np.sqrt(2.0) * 2.0, rho_max = 2**16,
                    max_iters = 20, max_inner_iters = 500, verbose=False)
        self.assertItemsAlmostEqual(X.value, true_X, places=2)

        # CG
        prob = Problem(prox_fns)
        prob.set_lin_solver("cg")
        sltn = prob.solve(solver="admm")
        self.assertItemsAlmostEqual(X.value, true_X, places=2)
        sltn = prob.solve(solver="hqs", eps_rel=1e-6,
                     rho_0 = 1.0, rho_scale = np.sqrt(2.0) * 2.0, rho_max = 2**16,
                    max_iters = 20, max_inner_iters = 500, verbose=False)
        self.assertItemsAlmostEqual(X.value, true_X, places=2)

        # Quad funcs.
        prob = Problem(prox_fns)
        sltn = prob.solve(solver="admm")
        self.assertItemsAlmostEqual(X.value, true_X, places=2)

        # Absorbing lin ops.
        prox_fns = [norm1(5*mul_elemwise(B,X)), sum_squares(-2*X, b=B)]
        prob = Problem(prox_fns)
        prob.set_absorb(True)
        prob.solve(solver="admm", eps_rel=1e-6, eps_abs=1e-6)

        cvx_X = cvx.Variable(4, 2)
        cost = cvx.sum_squares(-2*cvx_X - B) + cvx.norm(5*cvx.mul_elemwise(B,cvx_X), 1)
        cvx_prob = cvx.Problem(cvx.Minimize(cost))
        cvx_prob.solve(solver=cvx.CVXOPT)
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)

        prob.set_absorb(False)
        prob.solve(solver="admm", eps_rel=1e-6, eps_abs=1e-6)
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)

        # Constant offsets.
        prox_fns = [norm1(5*mul_elemwise(B,X)), sum_squares(-2*X - B)]
        prob = Problem(prox_fns)
        prob.solve(solver="admm", eps_rel=1e-6, eps_abs=1e-6)
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)



from .lin_op import LinOp
import numpy as np
import copy

class hstack(LinOp):
    """Horizontally concatenates vector inputs.
    """
    def __init__(self, input_nodes, implem=None):
        height = input_nodes[0].size
        width = len(input_nodes)
        super(hstack, self).__init__(input_nodes, (height,width))

    def forward(self, inputs, outputs):
        """The forward operator.

        Reads from inputs and writes to outputs.
        """
        offset = 0
        for idx, input_data in enumerate(inputs):
            outputs[0][:,idx] = input_data.flatten()

    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """
        for idx, output_data in enumerate(outputs):
            size = output_data.size
            data = inputs[0][:,idx]
            output_data[:] = np.reshape(data, output_data.shape)

    def is_gram_diag(self, freq=False):
        """Is the lin op's Gram matrix diagonal (in the frequency domain)?
        """
        return all([arg.is_gram_diag(freq) for arg in self.input_nodes])

    def get_diag(self, freq=False):
        """Returns the diagonal representation (A^TA)^(1/2).

        Parameters
        ----------
        freq : bool
            Is the diagonal representation in the frequency domain?
        Returns
        -------
        dict of variable to ndarray
            The diagonal operator acting on each variable.
        """
        var_diags = {var:np.zeros(var.size) for var in self.variables()}
        for arg in self.input_nodes:
            arg_diags = arg.get_diag(freq)
            for var, diag in arg_diags.items():
                var_diags[var] = var_diags[var]+diag*np.conj(diag)
        # Get (A^TA)^{1/2}
        for var in self.variables():
            var_diags[var] = np.sqrt(var_diags[var])
        return var_diags

    def norm_bound(self, input_mags):
        """Gives an upper bound on the magnitudes of the outputs given inputs.

        Parameters
        ----------
        input_mags : list
            List of magnitudes of inputs.

        Returns
        -------
        float
            Magnitude of outputs.
        """
        return np.linalg.norm(input_mags, 2)
